/*
 *  TOPPERS ATK2
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *      Automotive Kernel Version 2
 *
 *  Copyright (C) 2008-2017 by Center for Embedded Computing Systems
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2012-2017 by FUJI SOFT INCORPORATED, JAPAN
 *  Copyright (C) 2012-2013 by Spansion LLC, USA
 *  Copyright (C) 2012-2017 by NEC Communication Systems, Ltd., JAPAN
 *  Copyright (C) 2012-2016 by Panasonic Advanced Technology Development Co., Ltd., JAPAN
 *  Copyright (C) 2012-2014 by Renesas Electronics Corporation, JAPAN
 *  Copyright (C) 2012-2016 by Sunny Giken Inc., JAPAN
 *  Copyright (C) 2012-2017 by TOSHIBA CORPORATION, JAPAN
 *  Copyright (C) 2012-2017 by Witz Corporation
 *  Copyright (C) 2014-2016 by AISIN COMCRUISE Co., Ltd., JAPAN
 *  Copyright (C) 2014-2016 by eSOL Co.,Ltd., JAPAN
 *  Copyright (C) 2014-2017 by SCSK Corporation, JAPAN
 *  Copyright (C) 2015-2017 by SUZUKI MOTOR CORPORATION
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，AUTOSAR（AUTomotive Open System ARchitecture）仕
 *  様に基づいている．上記の許諾は，AUTOSARの知的財産権を許諾するもので
 *  はない．AUTOSARは，AUTOSAR仕様に基づいたソフトウェアを商用目的で利
 *  用する者に対して，AUTOSARパートナーになることを求めている．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  $Id: target_ici.h 2401 2017-03-14 09:09:24Z witz-itoyo $
 */

/*
 *  コア間割込みドライバ
 */

#ifndef TOPPERS_TARGET_ICI_H
#define TOPPERS_TARGET_ICI_H

/*
 *  コア間割込みの割込み番号
 */
#define INTNO_ICI0	(UINT_C(0x10000) | CORE_INT_0_INT)
#define INTNO_ICI1	(UINT_C(0x20000) | CORE_INT_1_INT)
#define INTNO_ICI2	(UINT_C(0x30000) | CORE_INT_2_INT)
#define INTNO_ICI3	(UINT_C(0x40000) | CORE_INT_3_INT)

/*
 *  コア間割込みの割込み優先度
 */
#ifndef INTPRI_ICI0
#define INTPRI_ICI0	3
#endif /* INTPRI_ICI0 */

#ifndef INTPRI_ICI1
#define INTPRI_ICI1	3
#endif /* INTPRI_ICI1 */

#ifndef INTPRI_ICI2
#define INTPRI_ICI2	3
#endif /* INTPRI_ICI2 */

#ifndef INTPRI_ICI3
#define INTPRI_ICI3	3
#endif /* INTPRI_ICI3 */

/*
 *  コア間割込み番号と優先度取得用マクロ
 */
#define GET_ICI_INTNO(coreid)	target_ici_intno_table[coreid]
#define GET_ICI_INTPRI(coreid)	target_ici_intpri_table[coreid]

#ifndef TOPPERS_MACRO_ONLY
/*
 *  コア間割込み番号
 */
extern const InterruptNumberType	target_ici_intno_table[];

/*
 *  コア間割込み優先度
 */
extern const PriorityType			target_ici_intpri_table[];
#endif /* TOPPERS_MACRO_ONLY */

/*
 *  プロセッサ依存モジュール（NIOS2用）
 */
#include "nios2_gcc/prc_ici.h"

#endif /* TOPPERS_TARGET_ICI_H */
